import logging
from django.core.management.base import BaseCommand, CommandError


class Command(BaseCommand):
    @property
    def logger(self):
        if not hasattr(self, '_logger'):
            self._logger = logging.getLogger('dbcron')
        return self._logger



from django.test import TestCase
from django.core.validators import ValidationError
from dbcron import validators
from dbcron.tests import factories


class ValidatorTestMixin:
    def test_integer(self):
        value = factories.faker.random_int(self.validator.int_min,
                                           self.validator.int_max)
        self.validator()(value)

    def test_range(self):
        delta = self.validator.int_max - self.validator.int_min
        min_ = factories.faker.random_int(
            max([1, self.validator.int_min]),
            int(self.validator.int_min+delta/2)
        )
        max_ = factories.faker.random_int(min_+1, self.validator.int_max)
        value = '%d-%d' % (min_, max_)
        self.validator()(value)

    def test_list(self):
        values = [
            factories.faker.random_int(self.validator.int_min,
                                       self.validator.int_max)
            for i in range(3)
        ]
        value = ','.join([str(i) for i in values])
        self.validator()(value)

    def test_frequency(self):
        value = '*/%d' % factories.faker.random_int(
            max([1, self.validator.int_min]),
            self.validator.int_max
        )
        self.validator()(value)

    def test_star(self):
        value = '*'
        self.validator()(value)

    def test_ranges_in_list(self):
        values = []
        delta = self.validator.int_max - self.validator.int_min
        for i in range(3):
            min_ = factories.faker.random_int(
                max([1, self.validator.int_min]),
                int(self.validator.int_min+delta/2)
            )
            max_ = factories.faker.random_int(min_+1, self.validator.int_max)
            values.append('%d-%d' % (min_, max_))
        value = ','.join(values)
        self.validator()(value)

    def test_too_small_integer(self):
        value = self.validator.int_min - 1
        with self.assertRaises(ValidationError):
            self.validator()(value)

    def test_too_big_integer(self):
        value = self.validator.int_max + 1
        with self.assertRaises(ValidationError):
            self.validator()(value)

    def test_invalid_integer_in_list(self):
        values = []
        values.append(factories.faker.random_int(self.validator.int_min,
                                                 self.validator.int_max))
        values.append(self.validator.int_max + 1)
        value = ','.join([str(i) for i in values])
        with self.assertRaises(ValidationError):
            self.validator()(value)

    def test_invalid_frequency(self):
        value = '*/-1'
        with self.assertRaises(ValidationError):
            self.validator()(value)

    def test_invalid_string(self):
        value = factories.faker.first_name()
        with self.assertRaises(ValidationError):
            self.validator()(value)

    def test_invalid_range_format(self):
        values = [
            factories.faker.random_int(self.validator.int_min,
                                       self.validator.int_max)
            for i in range(3)
        ]
        value = '-'.join([str(i) for i in values])
        with self.assertRaises(ValidationError):
            self.validator()(value)


class SecondsValidatorTest(ValidatorTestMixin, TestCase):
    validator = validators.SecondsValidator


class MinutesValidatorTest(ValidatorTestMixin, TestCase):
    validator = validators.MinutesValidator


class HoursValidatorTest(ValidatorTestMixin, TestCase):
    validator = validators.HoursValidator


class DaysOfMonthValidatorTest(ValidatorTestMixin, TestCase):
    validator = validators.DaysOfMonthValidator

    def test_L(self):
        value = 'L'
        self.validator()(value)

    def test_interrogation_mark(self):
        value = '?'
        self.validator()(value)


class MonthValidatorTest(ValidatorTestMixin, TestCase):
    validator = validators.MonthValidator

    def test_string_month(self):
        value = factories.faker.random_element(self.validator.month_list)
        self.validator()(value)


class DaysOfWeekValidatorTest(ValidatorTestMixin, TestCase):
    validator = validators.DaysOfWeekValidator

    def test_L(self):
        value = 'L'
        self.validator()(value)

    def test_string_day(self):
        value = factories.faker.random_element(self.validator.day_list)
        self.validator()(value)

    def test_string_day_lower(self):
        value = factories.faker.random_element(self.validator.day_list).lower()
        self.validator()(value)

    def test_interrogation_mark(self):
        value = '?'
        self.validator()(value)


class YearsValidatorTest(ValidatorTestMixin, TestCase):
    validator = validators.YearsValidator



from django.core import validators
from django.utils.translation import ugettext_lazy as _
from django.utils.deconstruct import deconstructible

MONTH_MAP = {
    'JAN': 1,
    'FEB': 2,
    'MAR': 3,
    'APR': 4,
    'MAY': 5,
    'JUN': 6,
    'JUL': 7,
    'AUG': 8,
    'SEP': 9,
    'OCT': 10,
    'NOV': 11,
    'DEC': 12,
}
DAYS = {
    'SUN': 0,
    'MON': 1,
    'TUE': 2,
    'WED': 3,
    'THU': 4,
    'FRI': 5,
    'SAT': 6,
}


@deconstructible
class BaseCrontabValidator:
    code = 'invalid'
    int_message = _("Enter a number between %d and %d.")
    range_message = _("Please enter valid range from %d and %s.")
    freq_message = _("Enter a valid positive number.")
    special_strings = []

    def __eq__(self, other):
        return self.__class__ == other.__class__

    def validate_int(self, value, index):
        try:
            value = int(value)
        except TypeError:
            raise validators.ValidationError(
                message=self.int_message % (self.int_min, self.int_max),
                code=self.code
            )
        if not self.int_min <= value <= self.int_max:
            raise validators.ValidationError(
                message=self.int_message % (self.int_min, self.int_max),
                code=self.code
            )

    def validate_range(self, value, index):
        try:
            min_, max_ = value.split('-')
        except ValueError:
            raise validators.ValidationError(
                message=_("Bad range format"),
                code=self.code
            )
        self.validate_int(min_, index)
        self.validate_int(max_, index)
        min_, max_ = int(min_), int(max_)
        if min_ >= max_:
            raise validators.ValidationError(
                message=self.range_message % (self.int_min, self.int_max),
                code=self.code
            )

    def validate_frequency(self, value, index):
        star, freq = value.split('/')
        try:
            freq = int(freq)
        except TypeError:
            raise validators.ValidationError(message=self.freq_message,
                                             code=self.code)
        if freq <= 0:
            raise validators.ValidationError(message=self.freq_message,
                                             code=self.code)
        if star != '*':
            raise validators.ValidationError(
                message=_("The first character of frequency must be '*'."),
                code=self.code
            )

    def __call__(self, value):
        value = str(value)
        values = value.split(',')
        for i, value in enumerate(values):
            if value.isdigit():
                self.validate_int(value, i)
                continue
            elif value.startswith('-') and value[1:].isdigit():
                raise validators.ValidationError(
                    message=self.range_message % (self.int_min, self.int_max),
                    code=self.code
                )
            elif value.startswith('*/'):
                self.validate_frequency(value, i)
                continue
            elif '-' in value:
                self.validate_range(value, i)
                continue
            elif value.upper() in self.special_strings:
                continue
            elif value in ['*', '?']:
                continue
            raise validators.ValidationError(
                message=_("Enter a correct value."),
                code=self.code
            )


class SecondsValidator(BaseCrontabValidator):
    int_min = 0
    int_max = 59


class MinutesValidator(BaseCrontabValidator):
    int_min = 0
    int_max = 59


class HoursValidator(BaseCrontabValidator):
    int_min = 0
    int_max = 23


class DaysOfMonthValidator(BaseCrontabValidator):
    int_min = 1
    int_max = 31
    special_strings = ['L']


class MonthValidator(BaseCrontabValidator):
    int_min = 1
    int_max = 12
    month_list = MONTH_MAP
    special_strings = list(month_list)


class DaysOfWeekValidator(BaseCrontabValidator):
    int_min = 0
    int_max = 6
    day_list = DAYS
    special_strings = ['L'] + list(day_list)


class YearsValidator(BaseCrontabValidator):
    int_min = 1970
    int_max = 2099



from django.conf import settings


def get_setting(name, *default):
    full_name = 'DBCRON_%s' % name
    value = getattr(settings, full_name, *default)
    return value


MAX_WORKERS = get_setting('MAX_WORKERS', 8)

JOB_GROUP = get_setting('JOB_GROUP', 'dbcron')



from django.contrib import admin
from django.utils.translation import ugettext_lazy as _
from dbcron import models
from dbcron import utils


@admin.register(models.Job)
class JobAdmin(admin.ModelAdmin):
    list_display = ['name', 'tag', 'func', 'is_active', 'next_time', 'sec',
                    'min', 'hou', 'dom', 'mon', 'dow', 'yea']
    list_filter = ['is_active', 'func', 'tag']
    actions = ('make_disable', 'make_enable')
    ordering = ['name']
    fieldsets = (
        (_('Metadata'), {
            'classes': ('wide',),
            'fields': (
                ('name', 'tag', 'is_active'),
                'description',
            )
        }),
        (_("Operation"), {
            'classes': ('wide',),
            'fields': (
                'func',
                'args',
                'opts',
            )
        }),
        (_("Scheduling"), {
            'classes': ('wide',),
            'fields': (
                ('sec', 'min', 'hou', 'dom', 'mon', 'dow', 'yea'),
            )
        }),
    )

    def next_time(self, obj):
        return utils.get_next_time(obj) or '-'
    next_time.short_description = _("Next run")

    def make_disable(self, request, queryset):
        queryset.update(is_active=False)
    make_disable.short_description = _("Disable job(s)")

    def make_enable(self, request, queryset):
        queryset.update(is_active=True)
    make_enable.short_description = _("Enable job(s)")



from datetime import date, timedelta, datetime
from calendar import HTMLCalendar
from django.utils.translation import ugettext_lazy as _
from dateutil.relativedelta import relativedelta
from dbcron import models

DAYS = {
    6: _("Sunday"),
    0: _("Monday"),
    1: _("Tuesday"),
    2: _("Wednesday"),
    3: _("Thursday"),
    4: _("Friday"),
    5: _("Saturday"),
}


class JobCalendar(HTMLCalendar):
    table_class = ""
    month_table_class = ""
    day_outside_class = ""
    month_today_class = ""
    day_of_month_class = ""
    ul_class = ""
    active_job_class = ""
    disactive_job_class = ""

    def __init__(self, jobs, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.jobs = jobs

    def get_table_class(self):
        return self.table_class

    def get_job_class(self, job):
        return self.active_job_class if job.is_active else self.disactive_job_class

    def formatweekday(self, day):
        return '<th class="%s">%s</th>' % (self.cssclasses[day], self._format_weekday(day))

    def formatmonth(self, theyear, themonth, withyear=True):
        v = []
        month_weeks = self.monthdays2calendar(theyear, themonth)
        days_before = len([i for i in month_weeks[0] if i[0] == 0])
        dates = self.jobs.get_next_planned_by_hour(
            from_=datetime(theyear, themonth, 1) - timedelta(days=days_before),
            until=date(theyear, themonth, 1) + relativedelta(months=1) + timedelta(days=7)
        )
        a = v.append
        a('<table class="%s">' % (
            self.month_table_class,
        ))
        a('\n')
        a(self.formatmonthname(theyear, themonth, withyear=withyear))
        a('\n')
        a(self.formatweekheader())
        a('\n')
        for week_num, week in enumerate(month_weeks):
            a('<tr>')
            for day_id, weekday in week:
                day_klasses = []
                if day_id == 0:  # Out of month
                    if week_num == 0:
                        delta = len([i for i in week if i[1] >= weekday and not i[0]])
                        day = date(theyear, themonth, 1) - timedelta(days=delta)
                    elif week_num == len(month_weeks) - 1:
                        delta = len([i for i in week if i[1] <= weekday and not i[0]])
                        day = self.get_lastdateofmonth(theyear, themonth) + timedelta(days=delta)
                    day_klasses.append(self.day_outside_class)
                else:
                    day = date(theyear, themonth, day_id)
                if day == date.today():
                    day_klasses.append(self.month_today_class)
                a('<td class="%s">' % ' '.join(day_klasses))
                a('<div class="%s">%s</div>' % (
                    self.day_of_month_class,
                    day.day
                ))
                self._format_month_jobs(v, day, dates)
                a('</td>')
            a('</tr>')
            a('\n')
        a('</table>')
        a('\n')
        return ''.join(v)

    def itermonthweekdays(self, theyear, themonth, theweek):
        for day_ in self.itermonthdates(theyear, themonth):
            if day_.isocalendar()[1] != theweek:
                continue
            yield day_

    def get_firstdateofweek(self, theyear, theweek):
        year_date = date(theyear, 1, 1)
        week_date = year_date + relativedelta(weeks=theweek-1)
        first_day = week_date - timedelta(days=week_date.isocalendar()[2]-1)
        return first_day

    def get_lastdateofmonth(self, theyear, themonth):
        month_after = date(theyear, themonth, 1) + relativedelta(months=1)
        current_month = month_after - timedelta(days=1)
        return current_month

    def _format_weekday(self, day):
        return str(DAYS[day])

    def _format_hour(self, hour):
        if hour == 0:
            return str(_("Midnight"))
        elif hour == 12:
            return str(_("Noon"))
        return '%d' % hour

    def _format_month_day_job(self, v, job, jobtime):
        a = v.append
        a('<li class="%s">' % self.get_job_class(job))
        if hasattr(job, 'get_absolute_url'):
            a('%s - <a href="%s">%s</a>' % (jobtime.strftime("%H:%M"),
                                            job.get_absolute_url(),
                                            job.name))
        else:
            a('%s - %s' % (jobtime.strftime("%H:%M"), job.name))
        a('</li>')

    def _format_month_jobs(self, v, day, dates):
        a = v.append
        a('<ul class="%s">' % self.ul_class)
        offset = date(day.year, day.month, day.day)
        for hour in dates[offset]:
            for job, jobtime in dates[offset][hour]:
                self._format_month_day_job(v, job, jobtime)
            offset += timedelta(days=1)
            a('\n')
        a('</ul>')

    def _format_day_job(self, v, job, jobtime):
        a = v.append
        if hasattr(job, 'get_absolute_url'):
            a('%s - <a href="%s">%s</a>' % (jobtime.strftime("%M"),
                                            job.get_absolute_url(),
                                            job.name))
        else:
            a('%s - %s' % (jobtime.strftime("%H:%M"), job.name))

    def _format_day_jobs(self, v, first_day, dates):
        a = v.append
        for hour in range(0, 24):
            day = first_day
            a('<tr>')
            a('<th>')
            a(self._format_hour(hour))
            a('</th>')
            for i in range(7):
                a('<td>')
                a('<ul class="%s">' % self.ul_class)
                for job, jobtime in dates[day][hour]:
                    a('<li class="%s">' % self.get_job_class(job))
                    self._format_day_job(v, job, jobtime)
                    a('</li>')
                a('</ul>')
                a('</td>')
                day += timedelta(days=1)
            a('</tr>')
            a('\n')

    def _format_weekdays_header(self, v, start_day):
        a = v.append
        day = start_day
        for i in range(7):
            a('<th>%s %s</th>' % (
                self._format_weekday(i),
                day.day,
            ))
            day += timedelta(days=1)

    def formatweekofmonth(self, theyear, theweek, withyear=True):
        v = []
        a = v.append
        day = self.get_firstdateofweek(theyear, theweek)
        a('<table class="%s">' % (
            self.get_table_class()
        ))
        a('\n')
        a('<tr>')
        a('<th>%s</th>' % _("UTC"))
        self._format_weekdays_header(v, day)
        a('</tr>')
        dates = self.jobs.get_next_planned_by_hour(
            from_=datetime(theyear, day.month, day.day) - timedelta(days=1),
            until=day+timedelta(days=7))
        a('<tr>')
        self._format_day_jobs(v, day, dates)
        a('<tr>')
        a('</table>')
        a('\n')
        return ''.join(v)



from faker import Faker
import factory
from factory import fuzzy

faker = Faker()


class Fuzzy(fuzzy.FuzzyInteger):
    def __init__(self, low=None, high=None, *args, **kwargs):
        super().__init__(low=self.low, high=self.high, *args, **kwargs)

    def fuzz(self):
        value = super().fuzz()
        return str(value)


class FuzzyMinute(Fuzzy):
    low = 0
    high = 59


class FuzzyHour(Fuzzy):
    low = 0
    high = 23


class JobFactory(factory.django.DjangoModelFactory):
    name = factory.Faker('job')
    tag = factory.Faker('name')
    func = 'django.utils.timezone.now'

    sec = '0'
    min = FuzzyMinute(0, 59)
    hou = FuzzyHour(0, 23)
    dom = '*'
    mon = '*'
    dow = '*'
    yea = '*'

    class Meta:
        model = 'dbcron.Job'



"""Manage your crontab directly from DB."""
VERSION = (0, 5, 3)
__version__ = '.'.join([str(i) for i in VERSION])
__author__ = 'Anthony Monthe (ZuluPro)'
__email__ = 'amonthe@cloudspectator.com'
__url__ = 'https://github.com/cloudspectatordevelopment/django-dbcron'
default_app_config = 'dbcron.apps.DbcronConfig'



from unittest.mock import patch
from datetime import datetime
from io import StringIO

from django.test import TestCase
from django.core.management import call_command

from dbcron.management.commands.crond import Command as CrondCommand
from dbcron.management.commands.run_jobs import Command as RunJobsCommand
from dbcron.tests.factories import JobFactory


class CrondRunJobTest(TestCase):
    def setUp(self):
        self.command = CrondCommand()

    @patch('crontab.CronTab.next', return_value=42)
    def test_not_now(self, mock):
        job = JobFactory.create()
        result = self.command.run_job(job)
        self.assertIsNone(result)

    @patch('crontab.CronTab.next', return_value=0)
    def test_run(self, mock):
        job = JobFactory.create()
        result = self.command.run_job(job)
        self.assertIsNotNone(result)
        self.assertIsInstance(result, datetime)

    @patch('crontab.CronTab.next', side_effect=Exception)
    def test_run_with_error(self, mock):
        job = JobFactory.create()
        with self.assertRaises(Exception):
            self.command.run_job(job)


class CrondCommandTest(TestCase):
    def setUp(self):
        self.stdout = StringIO()

    @patch('dbcron.management.commands.crond.time.sleep')
    @patch('dbcron.management.commands.crond.Command.stop', side_effect=(False, True))
    def test_run(self, mock, mock_sleep):
        call_command('crond', stdout=self.stdout)
        self.assertEqual(mock.call_count, 2)
        self.stdout.seek(0)
        self.assertIn('Started', self.stdout.read())

    @patch('dbcron.management.commands.crond.time.sleep')
    @patch('dbcron.management.commands.crond.Command.run_job', return_value=True)
    @patch('dbcron.management.commands.crond.Command.stop', return_value=True)
    def test_filter_tag(self, mock_stop, mock_run_job, mock_sleep):
        JobFactory.create(tag='foo', is_active=True)
        call_command('crond', '--tags', 'foo', stdout=self.stdout)
        self.assertEqual(mock_run_job.call_count, 1)

    @patch('dbcron.management.commands.crond.time.sleep')
    @patch('dbcron.management.commands.crond.Command.run_job', return_value=True)
    @patch('dbcron.management.commands.crond.Command.stop', return_value=True)
    def test_filter_tag_has_no_job(self, mock_stop, mock_run_job, mock_sleep):
        JobFactory.create(tag='foo', is_active=True)
        call_command('crond', '--tags', 'bar', stdout=self.stdout)
        self.assertEqual(mock_run_job.call_count, 0)

    @patch('dbcron.management.commands.crond.time.sleep')
    @patch('dbcron.management.commands.crond.Command.run_job', return_value=True)
    @patch('dbcron.management.commands.crond.Command.stop', return_value=True)
    def test_filter_tags(self, mock_stop, mock_run_job, mock_sleep):
        JobFactory.create(tag='foo', is_active=True)
        JobFactory.create(tag='bar', is_active=True)
        JobFactory.create(tag='ham', is_active=True)
        call_command('crond', '--tags', 'foo', 'bar', stdout=self.stdout)
        self.assertEqual(mock_run_job.call_count, 2)

    @patch('dbcron.management.commands.crond.time.sleep')
    @patch('dbcron.management.commands.crond.Command.stop', return_value=True)
    def test_quiet(self, mock_stop, mock_sleep):
        call_command('crond', '--quiet', stdout=self.stdout)
        self.stdout.seek(0)
        self.assertFalse(self.stdout.read())

    @patch('dbcron.management.commands.crond.Command.main', side_effect=KeyboardInterrupt)
    def test_keyboard_interrupt(self, mock_main):
        call_command('crond', stdout=self.stdout)
        self.stdout.seek(0)
        self.assertIn('Stopped', self.stdout.read())


class RunJobsCommandTest(TestCase):
    def setUp(self):
        self.stdout = StringIO()

    def test_run_one(self):
        job = JobFactory.create()
        call_command('run_jobs', job.id, stdout=self.stdout)
        self.stdout.seek(0)
        self.assertIn('Started', self.stdout.read())

    def test_run_two(self):
        job1, job2 = JobFactory.create_batch(2)
        call_command('run_jobs', job1.id, job2.id, stdout=self.stdout)
        self.stdout.seek(0)
        self.assertIn('Started', self.stdout.read())



from collections import defaultdict
from django.db import models


class JobQuerySet(models.QuerySet):
    def get_next_planned_by_day(self, until, from_=None):
        dates = defaultdict(list)
        for job in self.all():
            for next_date in job.get_next_planned(until, from_):
                dates[next_date.date()].append((job, next_date))
        for date in dates:
            dates[date] = sorted(dates[date], key=lambda x: x[1].hour)
        return dates

    def get_next_planned_by_hour(self, until, from_=None):
        dates = defaultdict(lambda: defaultdict(list))
        for job in self.all():
            for next_date in job.get_next_planned(until, from_):
                dates[next_date.date()][next_date.hour].append((job, next_date))
        for date in dates:
            for hour in dates[date]:
                dates[date][hour] = sorted(dates[date][hour], key=lambda x: x[1])
        return dates



import importlib
import json
from datetime import timedelta

from crontab import CronTab

from django.db import models
from django.utils.timezone import now
from django.utils.translation import ugettext_lazy as _

from dbcron import validators
from dbcron import querysets
from dbcron import signals


class Job(models.Model):
    name = models.CharField(max_length=200, verbose_name=_("name"))
    tag = models.CharField(max_length=150, verbose_name=_("tag"), blank=True, null=True)
    description = models.TextField(max_length=2000, verbose_name=_("description"), blank=True)
    func = models.CharField(max_length=250, verbose_name=_("function"))

    args = models.TextField(max_length=2000, default='[]', verbose_name=_("arguments"))
    opts = models.TextField(max_length=2000, default='{}', verbose_name=_("options"))

    is_active = models.BooleanField(default=False, verbose_name=_("is active"))

    sec = models.CharField(
        max_length=50, default='0', verbose_name=_("second(s)"),
        validators=[validators.SecondsValidator()]
    )
    min = models.CharField(
        max_length=50, verbose_name=_("minute(s)"),
        validators=[validators.MinutesValidator()]
    )
    hou = models.CharField(
        max_length=50, verbose_name=_("hour(s)"),
        validators=[validators.HoursValidator()]
    )
    dom = models.CharField(
        max_length=50, verbose_name=_("day(s) of month"),
        validators=[validators.DaysOfMonthValidator()]
    )
    mon = models.CharField(
        max_length=50, verbose_name=_("month"),
        validators=[validators.MonthValidator()]
    )
    dow = models.CharField(
        max_length=50, verbose_name=_("day(s) of week"),
        validators=[validators.DaysOfWeekValidator()]
    )
    yea = models.CharField(
        max_length=50, default='*', verbose_name=_("year(s)"),
        validators=[validators.YearsValidator()]
    )

    objects = querysets.JobQuerySet.as_manager()

    class Meta:
        verbose_name = _("job")
        verbose_name_plural = _("jobs")

    @property
    def raw_entry(self):
        return ' '.join([self.sec, self.min, self.hou, self.dom, self.mon,
                         self.dow, self.yea])

    def __str__(self):
        return "%s (%s)" % (self.name, self.raw_entry)

    @property
    def entry(self):
        return CronTab(self.raw_entry.strip())

    def _get_func(self):
        module_path = '.'.join(self.func.split('.')[:-1])
        func_name = self.func.split('.')[-1]
        module = importlib.import_module(module_path)
        func = getattr(module, func_name)
        return func

    def get_arguments(self):
        return json.loads(self.args)

    def get_options(self):
        return json.loads(self.opts)

    def run(self, fail_silently=False):
        func = self._get_func()
        args = self.get_arguments()
        opts = self.get_options()
        signals.job_started.send(sender=self.__class__, job=self)
        try:
            result = func(*args, **opts)
        except Exception as err:
            signals.job_failed.send(sender=self.__class__, job=self, error=err)
            if fail_silently:
                return
            raise
        signals.job_done.send(sender=self.__class__, job=self)
        return result

    @property
    def next_time(self):
        next_ = self.entry.next()
        if next_ is not None:
            return int(self.entry.next())

    def get_next_planned(self, until, from_):
        from_ = from_ or now()
        while until > from_.date():
            delta = self.entry.next(from_)
            from_ += timedelta(seconds=delta)
            if delta < 1:
                continue
            yield from_



from datetime import timedelta
from django.template.defaultfilters import timeuntil
from django.utils.timezone import now


def get_next_time(obj):
    """Get humanized form of next run."""
    next_ = obj.next_time
    if next_ is None:
        return
    dst_date = now() + timedelta(seconds=next_)
    return timeuntil(dst_date)



# Generated by Django 2.0.8 on 2019-02-07 00:53

import dbcron.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Job',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=200, verbose_name='name')),
                ('tag', models.CharField(blank=True, max_length=150, null=True, verbose_name='tag')),
                ('description', models.TextField(blank=True, max_length=2000, verbose_name='description')),
                ('func', models.CharField(max_length=250, verbose_name='function')),
                ('args', models.TextField(default='[]', max_length=2000, verbose_name='arguments')),
                ('opts', models.TextField(default='{}', max_length=2000, verbose_name='options')),
                ('is_active', models.BooleanField(default=False, verbose_name='is active')),
                ('sec', models.CharField(default='0', max_length=50, validators=[dbcron.validators.SecondsValidator()], verbose_name='second(s)')),
                ('min', models.CharField(max_length=50, validators=[dbcron.validators.MinutesValidator()], verbose_name='minute(s)')),
                ('hou', models.CharField(max_length=50, validators=[dbcron.validators.HoursValidator()], verbose_name='hour(s)')),
                ('dom', models.CharField(max_length=50, validators=[dbcron.validators.DaysOfMonthValidator()], verbose_name='day(s) of month')),
                ('mon', models.CharField(max_length=50, validators=[dbcron.validators.MonthValidator()], verbose_name='month')),
                ('dow', models.CharField(max_length=50, validators=[dbcron.validators.DaysOfWeekValidator()], verbose_name='day(s) of week')),
                ('yea', models.CharField(default='*', max_length=50, validators=[dbcron.validators.YearsValidator()], verbose_name='year(s)')),
            ],
            options={
                'verbose_name_plural': 'jobs',
                'verbose_name': 'job',
            },
        ),
    ]



from django import forms
from django.utils.translation import ugettext_lazy as _
from dbcron import models


class AbstractJobForm:
    sec = forms.CharField(
        initial='0',
        label=_("Second(s)")
    )
    min = forms.CharField(
        initial='*',
        label=_("Minute(s)")
    )
    hou = forms.CharField(
        initial='*',
        label=_("Hour(s)")
    )
    dom = forms.CharField(
        initial='*',
        label=_("Day(s) of month")
    )
    mon = forms.CharField(
        initial='*',
        label=_("Month")
    )
    dow = forms.CharField(
        initial='*',
        label=_("Day(s) of week")
    )
    yea = forms.CharField(
        initial='*',
        label=_("Year(s)")
    )


class JobForm(AbstractJobForm, forms.ModelForm):
    class Meta:
        model = models.Job
        fields = '__all__'



import traceback

from django.dispatch import receiver

from asgiref.sync import async_to_sync, sync_to_async
from channels.generic import websocket
from channels.exceptions import StopConsumer
from channels import layers

from dbcron import models
from dbcron import signals
from dbcron import settings


class JobConsumerMixin:
    model = models.Job
    queryset = model.objects.all()
    groups = [settings.JOB_GROUP]

    def base_run(self, event):
        self.job_id = event['id']
        self.job = self.queryset.get(pk=self.job_id)
        self.job.run()
        raise StopConsumer()

    def _format_status(self, event):
        status = event.copy()
        status.pop('type', None)
        return status


class JobSyncConsumer(JobConsumerMixin, websocket.JsonWebsocketConsumer):
    def run(self, event):
        self.base_run(event)

    def emit_job_status(self, event):
        status = self._format_status(event)
        self.send_json(status)


class JobAsyncConsumer(JobConsumerMixin, websocket.AsyncJsonWebsocketConsumer):
    async def run(self, event):
        await sync_to_async(self.base_run)(event)

    async def emit_job_status(self, event):
        status = self._format_status(event)
        await self.send_json(status)


def _emit_job_status(job, status, message=None):
    data = {
        'id': job.id,
        'type': 'emit.job.status',
        'status': status,
        'message': message,
    }
    if hasattr(job, 'get_absolute_url'):
        data['url'] = str(job.get_absolute_url())
    channel_layer = layers.get_channel_layer()
    async_to_sync(channel_layer.group_send)(settings.JOB_GROUP, data)


@receiver(signals.job_started, dispatch_uid="ws_job_start")
def send_job_start(sender, job, **kwargs):
    _emit_job_status(job, 'started')


@receiver(signals.job_done, dispatch_uid="ws_job_don")
def send_job_done(sender, job, **kwargs):
    _emit_job_status(job, 'done')


@receiver(signals.job_failed, dispatch_uid="ws_job_failed")
def send_job_failed(sender, job, error, **kwargs):
    message = traceback.format_exc().splitlines()[-1]
    _emit_job_status(job, 'failed', message)



from django.test import TestCase
from dbcron.calendar import JobCalendar
from dbcron import models
from dbcron.tests.factories import JobFactory


class JobCalendarFormatMonthTest(TestCase):
    factory = JobFactory
    jobs = models.Job.objects.all()

    def test_meth(self):
        self.factory.create_batch(5, min=0, hou=0)
        calendar = JobCalendar(self.jobs)
        for year in range(2000, 2005):
            for month in range(1, 13):
                html = calendar.formatmonth(year, month)


class JobCalendarFormatWeekTest(TestCase):
    factory = JobFactory
    jobs = models.Job.objects.all()

    def test_meth(self):
        self.factory.create_batch(5, min=0, hou=0)
        calendar = JobCalendar(self.jobs)
        for year in range(2000, 2005):
            for week in range(1, 53):
                html = calendar.formatweekofmonth(year, week)



import time
from io import StringIO
from concurrent import futures

from dbcron.management.commands import _base
from dbcron import models
from dbcron import settings


class Command(_base.Command):
    def add_arguments(self, parser):
        parser.add_argument(
            '-t', '--tags', nargs='*',
            help='Filter jobs by tag(s)'
        )
        parser.add_argument(
            '-q', '--quiet', action='store_true',
            help='Disable any output (except logs)',
        )

    def stop(self):
        """Helper for tests"""
        return False

    def run_job(self, job):
        """Decides to run a job, launches it and handles errors."""
        # Time decision
        next_ = int(job.entry.next())
        if next_ != 0:
            self.logger.debug("%s will run in %ssec", job.name, next_)
            return
        # Run
        self.logger.info("started %s", job.name)
        try:
            result = job.run()
        except Exception as err:
            self.logger.exception("error with %s", job.name)
            raise
        finally:
            self.logger.info("finished %s", job.name)
        return result

    def main(self, executor, tags, **kwargs):
        """Infinite loop acting as cron daemon."""
        # Lazy filtering
        jobs = models.Job.objects.filter(is_active=True)
        if tags:
            jobs = jobs.filter(tag__in=tags)
        self.stdout.write(self.style.SUCCESS('Started'))
        # Main Loop
        while True:
            self.logger.debug("new loop")
            executor.map(self.run_job, jobs.all())
            if self.stop():
                break
            time.sleep(1)

    def handle(self, *args, **options):
        if options['quiet']:
            self.stdout = self.stdout = StringIO()
        executor = futures.ThreadPoolExecutor(max_workers=settings.MAX_WORKERS)
        try:
            self.main(executor, **options)
        except KeyboardInterrupt as err:
            executor.shutdown()
            self.stdout.write(self.style.WARNING('Stopped'))
            return



from django.utils.translation import ugettext_lazy as _

import django_tables2 as tables
from django_tables2 import columns

from dbcron import models
from dbcron import utils


class AbstractJobTable(tables.Table):
    next_time = columns.Column(
        orderable=False, verbose_name=_("Next run"))


    def render_next_time(self, value, record):
        return utils.get_next_time(record)


class JobTable(AbstractJobTable, tables.Table):
    class Meta:
        model = models.Job
        fields = sequence = ['name', 'tag', 'func', 'is_active', 'next_time',
                             'sec', 'min', 'hou', 'dom', 'mon', 'dow', 'yea']



from datetime import date
from django.views.generic.base import RedirectView
from django.urls import reverse_lazy as reverse
from django.utils.timezone import now
from django.utils.html import mark_safe
from dateutil.relativedelta import relativedelta
from dbcron import calendar


class JobCalendarMixin:
    calendar_class = calendar.JobCalendar

    def get_calendar_class(self):
        return self.calendar_class

    def get_calendar_kwargs(self):
        return {
            'jobs': self.object_list,
        }

    def get_calendar(self, calendar_class=None):
        klass = calendar_class or self.get_calendar_class()
        calendar_kwargs = self.get_calendar_kwargs()
        return klass(**calendar_kwargs)

    def get_context_data(self, *args, **kwargs):
        data = super().get_context_data(*args, **kwargs)
        calendar = self.get_calendar()
        data.update({
            'calendar': calendar,
        })
        return data


class JobMonthCalendarRedirectView(RedirectView):
    pattern_name = 'job-calendar-month'

    def get_redirect_url(self, *args, **kwargs):
        now_ = now()
        kwargs['year'] = now_.year
        kwargs['month'] = now_.month
        return super().get_redirect_url(*args, **kwargs)


class JobMonthCalendarMixin(JobCalendarMixin):
    def get_context_data(self, *args, **kwargs):
        data = super().get_context_data(*args, **kwargs)
        calendar_date = date(int(self.kwargs['year']), int(self.kwargs['month']), 1)
        html_calendar = data['calendar']
        month_calendar = html_calendar.formatmonth(calendar_date.year, calendar_date.month)
        today = date.today()
        today_url = reverse('job-calendar-month', kwargs={
            'year': today.year, 'month': today.month
        })
        is_current_calendar = bool(
            calendar_date.year == today.year and
            calendar_date.month == today.month
        )
        next_date = calendar_date + relativedelta(months=1)
        next_url = reverse('job-calendar-month', kwargs={
            'year': next_date.year, 'month': next_date.month
        })
        prev_date = calendar_date - relativedelta(months=1)
        prev_url = reverse('job-calendar-month', kwargs={
            'year': prev_date.year, 'month': prev_date.month
        })
        data.update({
            'calendar': mark_safe(month_calendar),
            'calendar_date': calendar_date,
            'calendar_type': 'monthly',
            'is_current_calendar': is_current_calendar,
            'today_url': today_url,
            'today_date': today,
            'next_url': next_url,
            'next_date': next_date,
            'prev_url': prev_url,
            'prev_date': prev_date,
        })
        return data


class JobWeekCalendarRedirectView(RedirectView):
    pattern_name = 'job-calendar-week'

    def get_redirect_url(self, *args, **kwargs):
        now_ = now()
        kwargs['year'] = now_.year
        kwargs['week'] = now_.date().isocalendar()[1]
        return super().get_redirect_url(*args, **kwargs)


class JobWeekCalendarMixin(JobCalendarMixin):
    def get_context_data(self, *args, **kwargs):
        data = super().get_context_data(*args, **kwargs)
        today = now().date()
        today_week = today.isocalendar()[1]
        week = int(self.kwargs.get('week', today_week))
        year = int(self.kwargs.get('year', today.year))
        html_calendar = data['calendar']
        calendar_date = date(year, 1, 1) + relativedelta(weeks=week-1)
        week_calendar = html_calendar.formatweekofmonth(year, week)
        if calendar_date.year == today.year and today_week == week:
            is_current_calendar = True
        else:
            is_current_calendar = False
        today_url = reverse('job-calendar-week', kwargs={
            'year': today.year, 'week': today_week,
        })
        next_date = calendar_date + relativedelta(weeks=1)
        next_url = reverse('job-calendar-week', kwargs={
            'year': next_date.year, 'week': next_date.isocalendar()[1]
        })
        prev_date = calendar_date - relativedelta(weeks=1)
        prev_url = reverse('job-calendar-week', kwargs={
            'year': prev_date.year, 'week': prev_date.isocalendar()[1]
        })
        data.update({
            'calendar': mark_safe(week_calendar),
            'calendar_date': calendar_date,
            'calendar_type': 'weekly',
            'today_url': today_url,
            'today_week': today_week,
            'is_current_calendar': is_current_calendar,
            'next_url': next_url,
            'next_date': next_date,
            'prev_url': prev_url,
            'prev_date': prev_date,
        })
        return data



#!/usr/bin/env python
from setuptools import setup, find_packages
import dbcron


def read_file(name):
    with open(name) as fd:
        return fd.read()

keywords = ['django', 'cron', 'crontab']

setup(
    name='django-dbcron',
    version=dbcron.__version__,
    description=dbcron.__doc__,
    long_description=read_file('README.rst'),
    author=dbcron.__author__,
    author_email=dbcron.__email__,
    install_requires=read_file('requirements.txt'),
    license='BSD',
    url=dbcron.__url__,
    keywords=keywords,
    packages=find_packages(exclude=[]),
    include_package_data=True,
    # test_suite='runtests.main',
    # tests_require=read_file('requirements-tests.txt'),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Environment :: Console',
        'Framework :: Django',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],
)



from django import forms
import django_filters as filters
from dbcron import models


def get_tags(request=None):
    return models.Job.objects.values_list('tag', 'tag').distinct()


class JobFilterSet(filters.FilterSet):
    id = filters.BaseInFilter(widget=forms.HiddenInput)
    is_active = filters.BooleanFilter()
    tags = filters.MultipleChoiceFilter(choices=get_tags)

    class Meta:
        model = models.Job
        fields = []



from io import StringIO
from dbcron.management.commands import _base
from dbcron import models


class Command(_base.Command):
    def add_arguments(self, parser):
        parser.add_argument(
            'job_ids', nargs='+',
            help="Job's IDs"
        )
        parser.add_argument(
            '-q', '--quiet', action='store_true',
            help='Disable any output (except logs)',
        )

    def handle(self, *args, **options):
        if options['quiet']:
            self.stdout = self.stdout = StringIO()
        jobs = models.Job.objects.filter(id__in=options['job_ids'])
        self.stdout.write(self.style.SUCCESS('Started run'))
        for job in jobs:
            self.stdout.write(self.style.SUCCESS('Started %s' % job.name))
            try:
                job.run()
                self.stdout.write(self.style.SUCCESS('Done %s' % job.name))
            except Exception as err:
                self.logger.exception("error with %s", job.name)
                continue
        self.stdout.write(self.style.SUCCESS('Done'))



from django.apps import AppConfig
from django.utils.translation import ugettext_lazy as _


class DbcronConfig(AppConfig):
    name = 'dbcron'
    label = 'dbcron'
    verbose_name = _("Job scheduling")

    def ready(self):
        from dbcron import signals  # noqa
        from dbcron import consumers  # noqa
